Require Import Metatheory.

Require Import Arith.
Require Import Decidable.
Require Import OrderedTypeEx.

Require Import var_map.
Require Import list_facts.
Require Import models.
Require Import queues.

Set Implicit Arguments.

Open Scope nat_scope.

Module VarPair := PairOrderedType Var_as_OT Var_as_OT.

Notation "p ==== q" := (VarPair.eq_dec p q) (at level 67).

Module Type BUFFER (L : LANGUAGE_MODEL).
  Parameter buffer : Type.
  Parameter empty : buffer.
  Parameter is_empty_queue : buffer -> tid -> loc -> Prop.

  (* Holds if the entire buffer has no values enqueued. *)
  Definition is_empty (B : buffer) := forall t r, is_empty_queue B t r.

  (* Holds if there are no values enqueued for a thread. *)
  Definition is_empty_tid B t := forall r, is_empty_queue B t r.

  (* Holds if there are no values enqueued for a thread. *)
  Definition is_empty_loc B r := forall t, is_empty_queue B t r.

  Parameter enqueue : buffer -> tid -> loc -> L.val -> buffer.
  Parameter dequeue : buffer -> tid -> loc -> option (L.val * buffer).
  Parameter size : buffer -> nat.
  Parameter for_all_queue : buffer -> (L.val -> Prop) -> tid -> loc -> Prop.
  Parameter top : buffer -> tid -> loc -> option L.val.

  Definition for_all_loc B P r := forall t, for_all_queue B P t r.

  Definition for_all_tid B P t := forall r, for_all_queue B P t r.

  Definition for_all B P := forall t r, for_all_queue B P t r.

  Axiom empty_is_empty : is_empty empty.
  Axiom is_empty_queue_dequeue : forall B t r t' r' v B',
    is_empty_queue B t r ->
    dequeue B t' r' = Some (v, B') ->
    is_empty_queue B' t r.
  Axiom is_empty_queue_dequeue_diff : forall B t r v B' t' r',
    dequeue B t r = Some (v, B') ->
    (~ VarPair.eq (t, r) (t', r')) ->
    (is_empty_queue B' t' r' <-> is_empty_queue B t' r').
  Axiom dequeue_not_is_empty_queue : forall B t r v B',
    dequeue B t r = Some (v, B') ->
    ~ is_empty_queue B t r.
  Axiom is_empty_queue_enqueue_diff : forall B t r v t' r',
    (~ VarPair.eq (t, r) (t', r')) ->
    (is_empty_queue B t' r' <-> is_empty_queue (enqueue B t r v) t' r').
  Axiom empty_size : forall B,
    is_empty B <-> size B = 0.
  Axiom dequeue_size : forall B t r v B',
    dequeue B t r = Some (v, B') ->
    size B = size B' + 1.
  Axiom dequeue_confluence : forall B t1 r1 v1 B1 t2 r2 v2 B2,
    dequeue B t1 r1 = Some (v1, B1) ->
    dequeue B t2 r2 = Some (v2, B2) ->
    (~ VarPair.eq (t1, r1) (t2, r2)) ->
    exists B1',
      dequeue B1 t2 r2 = Some (v2, B1') /\ dequeue B2 t1 r1 = Some (v1, B1').
  Axiom dequeue_enqueue : forall B t r v B' t' r' v',
    dequeue B t r = Some (v, B') ->
    dequeue (enqueue B t' r' v') t r = Some (v, enqueue B' t' r' v').
  Axiom dequeue_enqueue_empty : forall B t r v,
    is_empty B ->
    exists B', dequeue (enqueue B t r v) t r = Some (v, B')
      /\ is_empty B'.
  Axiom is_empty_queue_dec : forall B t r,
    decidable (is_empty_queue B t r).
  Axiom for_all_empty : forall B t r P,
    is_empty_queue B t r ->
    for_all_queue B P t r.
  Axiom for_all_generalize : forall B t r (P Q : L.val -> Prop),
    for_all_queue B P t r ->
    (forall v, P v -> Q v) ->
    for_all_queue B Q t r.
  Axiom for_all_enqueue_diff : forall B P t1 r1 t2 r2 v2,
    for_all_queue B P t1 r1 ->
    ~ VarPair.eq (t1, r1) (t2, r2) ->
    for_all_queue (enqueue B t2 r2 v2) P t1 r1.
  Axiom for_all_enqueue_same : forall B P t1 r1 v1,
    for_all_queue B P t1 r1 ->
    P v1 ->
    for_all_queue (enqueue B t1 r1 v1) P t1 r1.
  Axiom for_all_dequeue_val : forall B1 P t1 r1 v1 B2,
    for_all_queue B1 P t1 r1 ->
    dequeue B1 t1 r1 = Some (v1, B2) ->
    P v1.
  Axiom for_all_dequeue : forall B1 P t1 r1 t2 r2 v2 B2,
    for_all_queue B1 P t1 r1 ->
    dequeue B1 t2 r2 = Some (v2, B2) ->
    for_all_queue B2 P t1 r1.
  Axiom for_all_top : forall B P t r v,
    for_all_queue B P t r ->
    top B t r = Some v ->
    P v.
  Axiom empty_top : forall B t r,
    is_empty_queue B t r ->
    top B t r = None.
  Axiom top_dequeue : forall B t r v v' B',
    top B t r = Some v ->
    dequeue B t r = Some (v', B') ->
    (v = v' /\ is_empty_queue B' t r) \/ (top B' t r = Some v).
  Axiom top_dequeue_diff : forall B t r opt t' r' v' B',
    top B t r = opt ->
    dequeue B t' r' = Some (v', B') ->
    ~ VarPair.eq (t, r) (t', r') ->
    top B' t r = opt.
  Axiom top_not_empty : forall B t r v,
    top B t r = Some v ->
    ~ is_empty_queue B t r.
End BUFFER.
