Require Import Metatheory.

Require Import models.
Require Import types.
Require Import var_map.

Set Implicit Arguments.

Module Type DRF_LANGUAGE_AXIOMS.
  Declare Module L : LANGUAGE_MODEL.
  Declare Module Typ : TYPE_MODEL with Module L := L.
  Declare Module Pres : TYPE_MODEL_PRESERVATION
    with Module L := L with Module Typ := Typ.

(*   Axiom well_formed_alloc_read : forall Sigma P t1 r1 v1 P1 t2 v2 P2, *)
(*     Typ.typing Sigma P -> *)
(*     ~ Map.In r1 Sigma -> *)
(*     L.program_step P (Some (t1, alloc r1 v1)) P1 -> *)
(*     L.program_step P1 (Some (t2, read r1 v2)) P2 -> *)
(*     t1 = t2. *)

(*   Axiom well_formed_alloc_write : forall Sigma P t1 r1 v1 P1 t2 v2 P2, *)
(*     Typ.typing Sigma P -> *)
(*     ~ Map.In r1 Sigma -> *)
(*     L.program_step P (Some (t1, alloc r1 v1)) P1 -> *)
(*     L.program_step P1 (Some (t2, write r1 v2)) P2 -> *)
(*     t1 = t2. *)

  Axiom well_formed_read : forall Sigma P1 t r v P2,
    Typ.typing Sigma P1 ->
    L.program_step P1 (Some (t, read r v)) P2 ->
    Map.In r Sigma.

  Axiom well_formed_write : forall Sigma P1 t r v P2,
    Typ.typing Sigma P1 ->
    L.program_step P1 (Some (t, write r v)) P2 ->
    Map.In r Sigma.

  Inductive program_conflict :
    tid * action L.val -> tid * action L.val -> Prop :=
  | program_conflict_tid : forall t a1 a2,
    program_conflict (t, a1) (t, a2)
  | program_conflict_spawn_left : forall t1 t2 a2,
    program_conflict (t1, spawn _ t2) (t2, a2)
  | program_conflict_spawn_right : forall t1 a1 t2,
    program_conflict (t1, a1) (t2, spawn _ t1).

  Axiom program_asynchrony : forall P P1 P2 e1 e2,
    L.program_step P (Some e1) P1 ->
    L.program_step P1 (Some e2) P2 ->
    ~ program_conflict e1 e2 ->
    exists P1',
      L.program_step P (Some e2) P1'
      /\ L.program_step P1' (Some e1) P2.

  Axiom value_independent_reads : forall P P' t r v v',
    L.program_step P (Some (t, read r v)) P' ->
    exists P'', L.program_step P (Some (t, read r v')) P''.

  Axiom spawn_fresh : forall P1 P2 P3 tr t1 t2 j t3 a3,
    L.program_step_many P1 P2 tr ->
    L.program_step P2 (Some (t1, spawn _ t2)) P3 ->
    index tr j (Some (t3, a3)) ->
    t2 <> t3.
End DRF_LANGUAGE_AXIOMS.
