Require Import Metatheory.

Require Import maps.
Require Import models.
Require Import msi.
Require Import msi_owner.

Module MSIToMSIOwner (L : LANGUAGE_MODEL).
  Module MSIO := MSIOwner L.
  Module MSI := MSI L.
  Declare Module SysMSIO : SYSTEM_MODEL L MSIO.
  Declare Module SysMSI : SYSTEM_MODEL L MSI.

  Definition map_heap (H1 : MSI.heap) (H2 : MSIO.heap) :=
    let (S1, C1) := H1 in
      let S2 := H2 in
        forall r,
          (match VarMap.find r S2 with
             | None =>
               VarMap.find r S1 = None
             | Some (MSIO.shared tids, v) =>
               VarMap.find r S1 = Some v
             | Some (MSIO.exclusive t', v) =>
               match VarMap.find r S1 with
                 | None => False
                 | Some _ => True
               end
           end) /\
          (forall t,
            match VarMap.find r S2 with
              | None =>
                VarPairMap.find (t, r) C1 = None
              | Some (MSIO.shared tids, v) =>
                match VarMap.find t tids with
                  | None => VarPairMap.find (t, r) C1 = None
                  | Some _ => VarPairMap.find (t, r) C1 = Some (MSI.shared, v)
                end
              | Some (MSIO.exclusive t', v) =>
                VarPairMap.find (t, r) C1 =
                if t == t' then Some (MSI.modified, v) else None
            end).

  Lemma map_heap_empty : map_heap MSI.empty_heap MSIO.empty_heap.
  Proof.
    unfold MSI.empty_heap, MSIO.empty_heap.
    unfold map_heap.
    intros r.
    rewrite VarMapFacts.empty_o.
    rewrite VarMapFacts.empty_o.
    split. auto.
    intros t.
    rewrite VarPairMapFacts.empty_o.
    auto.
  Qed.

  Lemma map_heap_alloc : forall S1 C1 S2 r v,
    map_heap (S1, C1) S2 ->
    ~ VarMap.In r S2 ->
    map_heap (VarMap.add r v S1, C1)
    (VarMap.add r (MSIO.shared (VarMap.empty _), v) S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 r v MapHeap NotIn r'.
    specializes MapHeap (>>> r').
    rewrite VarMapFacts.not_find_in_iff in NotIn.
    destruct (r == r'). subst r'.
    rewrite NotIn in MapHeap.
    destruct MapHeap.
    rewrite VarMapFacts.add_eq_o...
    rewrite VarMapFacts.add_eq_o...
    split... intros t'.
    rewrite VarMapFacts.empty_o...
    rewrite VarMapFacts.add_neq_o...
    rewrite VarMapFacts.add_neq_o...
  Qed.

  Lemma tids_neq : forall t r t' r',
    t <> t' ->
    ~ VarPair.eq (t, r) (t', r').
  Proof.
    intros t r t' r' Neq [ ? ? ].
    simpl in *. subst. tauto.
  Qed.

  Lemma locs_neq : forall t r t' r',
    r <> r' ->
    ~ VarPair.eq (t, r) (t', r').
  Proof.
    intros t r t' r' Neq [ ? ? ].
    simpl in *. subst. tauto.
  Qed.

  Lemma map_heap_write_modified : forall S1 C1 S2 t r v v',
    map_heap (S1, C1) S2 ->
    VarMap.MapsTo r (MSIO.exclusive t, v) S2 ->
    map_heap (S1, VarPairMap.add (t, r) (MSI.modified, v') C1)
    (VarMap.add r (MSIO.exclusive t, v') S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 t r v v' MapHeap Exclusive2 r'.
    specializes MapHeap (>>> r').
    destruct MapHeap as [ MapStore MapCache ].
    apply VarMap.find_1 in Exclusive2.
    destruct (r == r'). subst r'.
    rewrite Exclusive2 in MapStore.
    rewrite VarMapFacts.add_eq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite Exclusive2 in MapCache.
    destruct (t' == t). subst t'.
    rewrite VarPairMapFacts.add_eq_o...
    rewrite VarPairMapFacts.add_neq_o...
    apply tids_neq...
    rewrite VarMapFacts.add_neq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite VarPairMapFacts.add_neq_o...
    apply locs_neq...
  Qed.

  Lemma map_heap_invalid_to_shared : forall S1 C1 S2 t r v tids,
    map_heap (S1, C1) S2 ->
    VarMap.MapsTo r (MSIO.shared tids, v) S2 ->
    ~ VarMap.In t tids ->
    map_heap (S1, VarPairMap.add (t, r) (MSI.shared, v) C1)
    (VarMap.add r (MSIO.shared (VarMap.add t tt tids), v) S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 t r v tids MapHeap Shared2 NotInTIDs r'.
    specializes MapHeap (>>> r').
    destruct MapHeap as [ MapStore MapCache ].
    apply VarMap.find_1 in Shared2.
    rewrite VarMapFacts.not_find_in_iff in NotInTIDs.
    destruct (r == r'). subst r'.
    rewrite Shared2 in MapStore.
    rewrite VarMapFacts.add_eq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite Shared2 in MapCache.
    destruct (t == t'). subst t'.
    rewrite NotInTIDs in MapCache.
    rewrite VarMapFacts.add_eq_o...
    rewrite VarPairMapFacts.add_eq_o...
    rewrite VarMapFacts.add_neq_o...
    rewrite VarPairMapFacts.add_neq_o...
    apply tids_neq...
    rewrite VarMapFacts.add_neq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite VarPairMapFacts.add_neq_o...
    apply locs_neq...
  Qed.

  Lemma map_heap_invalid_to_modified : forall S1 C1 S2 t r v tids,
    map_heap (S1, C1) S2 ->
    VarMap.MapsTo r (MSIO.shared tids, v) S2 ->
    VarMap.Empty tids ->
    map_heap (S1, VarPairMap.add (t, r) (MSI.modified, v) C1)
    (VarMap.add r (MSIO.exclusive t, v) S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 t r v tids MapHeap Shared Empty r'.
    specializes MapHeap (>>> r').
    destruct MapHeap as [ MapStore MapCache ].
    apply VarMap.find_1 in Shared.
    destruct (r == r'). subst r'.
    rewrite Shared in MapStore.
    rewrite VarMapFacts.add_eq_o...
    rewrite MapStore.
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite Shared in MapCache.
    destruct (t' == t). subst t'.
    rewrite VarPairMapFacts.add_eq_o...
    rewrite VarPairMapFacts.add_neq_o...
    forwards NotInTIDs : VarMapFactsEx2.empty_find (>>> Empty t').
    rewrite NotInTIDs in MapCache...
    apply tids_neq...
    rewrite VarMapFacts.add_neq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite VarPairMapFacts.add_neq_o...
    apply locs_neq...
  Qed.

  Lemma map_heap_shared_to_invalid : forall S1 C1 S2 t r v tids,
    map_heap (S1, C1) S2 ->
    VarMap.MapsTo r (MSIO.shared tids, v) S2 ->
    VarMap.In t tids ->
    map_heap (S1, VarPairMap.remove (t, r) C1)
    (VarMap.add r (MSIO.shared (VarMap.remove t tids), v) S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 t r v tids MapHeap Shared2 InTIDs r'.
    specializes MapHeap (>>> r').
    destruct MapHeap as [ MapStore MapCache ].
    apply VarMap.find_1 in Shared2.
    destruct InTIDs as [ ? InTIDs ].
    apply VarMap.find_1 in InTIDs.
    destruct (r == r'). subst r'.
    rewrite Shared2 in MapStore.
    rewrite VarMapFacts.add_eq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite Shared2 in MapCache.
    destruct (t == t'). subst t'.
    rewrite VarMapFacts.remove_eq_o...
    rewrite VarPairMapFacts.remove_eq_o...
    rewrite VarMapFacts.remove_neq_o...
    rewrite VarPairMapFacts.remove_neq_o...
    apply tids_neq...
    rewrite VarMapFacts.add_neq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite VarPairMapFacts.remove_neq_o...
    apply locs_neq...
  Qed.

  Lemma map_heap_modified_to_invalid : forall S1 C1 S2 t r v,
    map_heap (S1, C1) S2 ->
    VarMap.MapsTo r (MSIO.exclusive t, v) S2 ->
    map_heap (VarMap.add r v S1, VarPairMap.remove (t, r) C1)
    (VarMap.add r (MSIO.shared (VarMap.empty _), v) S2).
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 t r v MapHeap Exclusive r'.
    specializes MapHeap (>>> r').
    destruct MapHeap as [ MapStore MapCache ].
    apply VarMap.find_1 in Exclusive.
    destruct (r == r'). subst r'.
    rewrite Exclusive in MapStore.
    rewrite VarMapFacts.add_eq_o...
    rewrite VarMapFacts.add_eq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite Exclusive in MapCache.
    rewrite VarMapFacts.empty_o.
    destruct (t' == t). subst t'.
    rewrite VarPairMapFacts.remove_eq_o...
    rewrite VarPairMapFacts.remove_neq_o...
    apply tids_neq...
    rewrite VarMapFacts.add_neq_o...
    rewrite VarMapFacts.add_neq_o...
    split... intros t'.
    specializes MapCache (>>> t').
    rewrite VarPairMapFacts.remove_neq_o...
    apply locs_neq...
  Qed.

  Lemma map_heap_fresh_msi : forall S1 C1 S2 r,
    map_heap (S1, C1) S2 ->
    ~ VarMap.In r S1 ->
    ~ VarMap.In r S2.
  Proof with auto.
    unfold map_heap.
    intros S1 C1 S2 r MapHeap Fresh2.
    specializes MapHeap (>>> r).
    destruct MapHeap as [ MapStore ].
    rewrite VarMapFacts.not_find_in_iff in Fresh2.
    rewrite Fresh2 in MapStore.
    rewrite VarMapFacts.not_find_in_iff.
    destruct (VarMap.find r S2) as [ [ [ tids | t ] v ] | ].
    inversion MapStore.
    inversion MapStore.
    auto.
  Qed.

  Ltac specialize_map_heap MapHeap t r :=
    unfold map_heap in MapHeap;
    specializes MapHeap (>>> r);
    destruct MapHeap as [ MapStore MapCache ];
    specializes MapCache (>>> t).

  Lemma map_heap_shared_msi : forall S1 C1 S2 t r v,
    map_heap (S1, C1) S2 ->
    VarPairMap.MapsTo (t, r) (MSI.shared, v) C1 ->
    exists tids,
      VarMap.MapsTo r (MSIO.shared tids, v) S2 /\
      VarMap.In t tids.
  Proof with auto.
    intros S1 C1 S2 t r v MapHeap Shared.
    specialize_map_heap MapHeap t r.
    apply VarPairMap.find_1 in Shared.
    remember (VarMap.find r S2) as find_S2.
    symmetry in Heqfind_S2.
    destruct find_S2 as [ [ [ t' | tids ] v' ] | ].
    destruct (t == t').
    assert (Some (MSI.shared, v) = Some (MSI.modified, v')) as false.
    rewrite <- Shared. rewrite <- MapCache...
    inversion false.
    assert (Some (MSI.shared, v) = None) as false.
    rewrite <- Shared. rewrite <- MapCache...
    inversion false.
    exists tids.
    remember (VarMap.find t tids) as opt.
    symmetry in Heqopt.
    destruct opt.
    assert (Some (MSI.shared, v) = Some (MSI.shared, v')) as Eq.
    rewrite <- Shared. rewrite <- MapCache...
    inversion Eq. subst v'.
    split. apply VarMap.find_2...
    esplit. apply VarMap.find_2. eauto.
    assert (Some (MSI.shared, v) = None) as false.
    rewrite <- Shared. rewrite <- MapCache...
    inversion false.
    assert (Some (MSI.shared, v) = None) as false.
    rewrite <- Shared. rewrite <- MapCache...
    inversion false.
  Qed.

  Lemma map_heap_modified_msi : forall S1 C1 S2 t r v,
    map_heap (S1, C1) S2 ->
    VarPairMap.MapsTo (t, r) (MSI.modified, v) C1 ->
    VarMap.MapsTo r (MSIO.exclusive t, v) S2.
  Proof with auto.
    intros S1 C1 S2 t r v MapHeap Modified.
    specialize_map_heap MapHeap t r.
    apply VarPairMap.find_1 in Modified.
    remember (VarMap.find r S2) as find_S2.
    symmetry in Heqfind_S2.
    destruct find_S2 as [ [ [ t' | tids ] v' ] | ].
    destruct (t == t'). subst t'.
    assert (Some (MSI.modified, v) = Some (MSI.modified, v')) as Eq.
    rewrite <- Modified. rewrite <- MapCache...
    inversion Eq. subst v'.
    apply VarMap.find_2...
    assert (Some (MSI.modified, v) = None) as false.
    rewrite <- Modified. rewrite <- MapCache...
    inversion false.
    destruct (VarMap.find t tids).
    assert (Some (MSI.modified, v) = Some (MSI.shared, v')) as false.
    rewrite <- Modified. rewrite <- MapCache...
    inversion false.
    assert (Some (MSI.modified, v) = None) as false.
    rewrite <- Modified. rewrite <- MapCache...
    inversion false.
    assert (Some (MSI.modified, v) = None) as false.
    rewrite <- Modified. rewrite <- MapCache...
    inversion false.
  Qed.

  Lemma map_heap_not_modified_msi : forall S1 C1 S2 r v,
    map_heap (S1, C1) S2 ->
    (forall (t : var) (v' : L.val),
      ~ VarPairMap.MapsTo (t, r) (MSI.modified, v') C1) ->
    VarMap.MapsTo r v S1 ->
    exists tids, VarMap.MapsTo r (MSIO.shared tids, v) S2.
  Proof.
    unfold map_heap.
    intros S1 C1 S2 r v MapHeap NotModified Store1.
    specializes MapHeap (>>> r).
    destruct MapHeap as [ MapStore MapCache ].
    remember (VarMap.find r S2) as opt.
    symmetry in Heqopt.
    destruct opt as [ [ [ t | tids ] v' ] | ].
    specializes MapCache (>>> t).
    destruct (t == t).
    apply VarPairMap.find_2 in MapCache.
    specializes NotModified (>>> t v'). tauto. tauto.
    exists tids.
    apply VarMap.find_2 in MapStore.
    forwards : VarMapFacts.MapsTo_fun (>>> MapStore Store1). subst.
    apply VarMap.find_2. auto.
    apply VarMap.find_1 in Store1.
    rewrite Store1 in MapStore. inversion MapStore.
  Qed.

  Lemma map_heap_invalid_not_modified_msi : forall S1 C1 S2 t r v,
    map_heap (S1, C1) S2 ->
    ~ VarPairMap.In (t, r) C1 ->
    (forall (t' : var) (v' : L.val),
      t <> t' -> ~ VarPairMap.MapsTo (t', r) (MSI.modified, v') C1) ->
    VarMap.MapsTo r v S1 ->
    exists tids,
      VarMap.MapsTo r (MSIO.shared tids, v) S2 /\ ~ VarMap.In t tids.
  Proof.
    intros S1 C1 S2 t r v MapHeap Invalid1 NotModified Store1.
    forwards [ tids Shared2 ] : map_heap_not_modified_msi.
    apply MapHeap.
    intros t' v'.
    destruct (t == t'). subst t'.
    intro. apply Invalid1. esplit. eauto.
    eauto. eauto.
    exists tids. split. auto.
    specialize_map_heap MapHeap t r.
    apply VarMap.find_1 in Shared2.
    rewrite Shared2 in MapCache.
    intros [ ? InTIDs ].
    apply VarMap.find_1 in InTIDs.
    rewrite InTIDs in MapCache.
    apply Invalid1. esplit.
    apply VarPairMap.find_2. eauto.
  Qed.

  Lemma map_heap_invalid_msi : forall S1 C1 S2 r v,
    map_heap (S1, C1) S2 ->
    (forall t, ~ VarPairMap.In (t, r) C1) ->
    VarMap.MapsTo r v S1 ->
    exists tids,
      VarMap.MapsTo r (MSIO.shared tids, v) S2 /\ VarMap.Empty tids.
  Proof with eauto.
    intros S1 C1 S2 r v MapHeap Invalid1 Store1.
    forwards [ ? Shared2 ] : map_heap_not_modified_msi...
    intros t v' Modified1.
    specializes Invalid1 (>>> t).
    apply Invalid1. esplit. eauto.
    esplit. split. eauto.
    intros t ? InTIDs.
    specialize_map_heap MapHeap t r.
    apply VarMap.find_1 in Shared2.
    rewrite Shared2 in MapCache.
    apply VarMap.find_1 in InTIDs.
    rewrite InTIDs in MapCache.
    specializes Invalid1 (>>> t).
    apply Invalid1. esplit.
    apply VarPairMap.find_2. eauto.
  Qed.

  Lemma msi_to_msi_owner_step : forall H1 H1' H2 e,
    map_heap H1 H2 ->
    MSI.heap_step H1 e H1' ->
    exists H2',
      MSIO.heap_step H2 e H2' /\
      map_heap H1' H2'.
  Proof with eauto.
    intros H1 H1' S2 e MapHeap Step.
    destruct H1 as [ S1 C1 ].
    destruct H1' as [ S1' C1' ].
    inversion Step; subst.

    (* Alloc *)
    forwards : map_heap_fresh_msi...
    esplit. split.
    apply MSIO.step_alloc...
    apply map_heap_alloc...

    (* Read shared *)
    forwards [ ? [ ? ? ] ] : map_heap_shared_msi...
    esplit. split.
    eapply MSIO.step_read_shared...
    auto.

    (* Read modified *)
    forwards : map_heap_modified_msi...
    esplit. split.
    eapply MSIO.step_read_modified...
    auto.

    (* Write modified *)
    forwards : map_heap_modified_msi...
    esplit. split.
    eapply MSIO.step_write_modified...
    eapply map_heap_write_modified...

    (* Invalid to shared *)
    forwards [ ? [ ? ? ] ] : map_heap_invalid_not_modified_msi...
    esplit. split.
    apply MSIO.step_invalid_to_shared...
    apply map_heap_invalid_to_shared...

    (* Invalid to modified *)
    forwards [ ? [ ? ? ] ] : map_heap_invalid_msi...
    esplit. split.
    eapply MSIO.step_invalid_to_modified...
    eapply map_heap_invalid_to_modified...

    (* Shared to invalid *)
    forwards [ ? [ ? ? ] ] : map_heap_shared_msi...
    esplit. split.
    apply MSIO.step_shared_to_invalid...
    apply map_heap_shared_to_invalid...

    (* Modified to invalid *)
    forwards : map_heap_modified_msi...
    esplit. split.
    eapply MSIO.step_modified_to_invalid...
    apply map_heap_modified_to_invalid...

    (* Spawn *)
    esplit. split.
    apply MSIO.step_spawn.
    auto.
  Qed.

  Lemma msi_to_msi_owner_step_many : forall H1 H2 P H1' P' tr,
    map_heap H1 H2 ->
    SysMSI.system_step_many H1 P H1' P' tr ->
    exists H2',
      SysMSIO.system_step_many H2 P H2' P' tr
      /\ map_heap H1' H2'.
  Proof with eauto.
    intros H1 H2 P H1' P' tr MapHeap StepMany.
    induction StepMany as [ | ? ? ? ? ? ? ? ? StepMany ? Step ].

    esplit. split.
    apply SysMSIO.system_step_many_refl.
    auto.

    specializes IHStepMany (>>> __)...
    destruct IHStepMany as [ ? [ ? ? ] ].
    inversion Step; subst.

    forwards [ ? [ ? ? ] ] : msi_to_msi_owner_step...
    esplit. split.
    eapply SysMSIO.system_step_many_trans...
    apply SysMSIO.system_step_mutual...
    auto.

    forwards [ ? [ ? ? ] ] : msi_to_msi_owner_step...
    esplit. split.
    eapply SysMSIO.system_step_many_trans...
    apply SysMSIO.system_step_heap...
    auto.

    esplit. split...
    eapply SysMSIO.system_step_many_trans...
    apply SysMSIO.system_step_program...
  Qed.

  Theorem msi_to_msi_owner : forall P0 H1 P tr,
    SysMSI.system_step_many MSI.empty_heap P0 H1 P tr ->
    exists H2,
      SysMSIO.system_step_many MSIO.empty_heap P0 H2 P tr.
  Proof.
    intros P0 H1 P tr StepMany.
    forwards [ ? [ ? ? ] ] : msi_to_msi_owner_step_many; eauto.
    apply map_heap_empty.
  Qed.
End MSIToMSIOwner.