Require Import Metatheory.

Require Import locks.
Require Import models.
Require Import stores.
Require Import var_map.

Set Implicit Arguments.

Open Scope type_scope.

Module Sequential_Consistency (L : LANGUAGE_MODEL) <: MEMORY_MODEL L.
  Declare Module Store : STORE L.

  Definition lock_map := LockMap.lock_map.

  Definition heap := (Store.store * lock_map).

  Definition empty_heap : heap := (Store.empty, LockMap.empty).

  Inductive step : heap -> effect L.val -> heap -> Prop :=
  | step_alloc : forall S Lambda t r v,
    Store.fresh S r ->
    step (S, Lambda) (Some (t, alloc r v)) (Store.update S r v, Lambda)
  | step_read : forall S Lambda t r v,
    Store.get S r v ->
    step (S, Lambda) (Some (t, read r v)) (S, Lambda)
  | step_write : forall S Lambda t r v,
    step (S, Lambda) (Some (t, write r v)) (Store.update S r v, Lambda)
  | step_acquire : forall S Lambda t l,
    LockMap.available Lambda l ->
    step (S, Lambda) (Some (t, acquire _ l)) (S, LockMap.acquire Lambda l t)
  | step_release : forall S Lambda t l,
    LockMap.held_by Lambda l t ->
    step (S, Lambda) (Some (t, release _ l)) (S, LockMap.release Lambda l)
  | step_spawn : forall H t1 t2,
    step H (Some (t1, spawn _ t2)) H.

  Definition heap_step := step.

  Fact no_pure_steps : forall H H' t,
    ~ (heap_step H (Some (t, pure _)) H').
  Proof.
    intros H H' t steps.
    inversion steps.
  Qed.

  Inductive heap_step_many : heap -> heap -> trace L.val -> Prop :=
  | heap_step_many_refl : forall H,
    heap_step_many H H nil
  | heap_step_many_trans : forall H1 H2 H3 tr e,
    heap_step_many H1 H2 tr ->
    heap_step H2 e H3 ->
    heap_step_many H1 H3 (e :: tr).

  Hint Unfold heap_step : sequential_consistency.
  Hint Constructors step : sequential_consistency.
End Sequential_Consistency.
